import { ChangeDetectorRef, Component, ComponentFactoryResolver, ContentChild, ElementRef, EventEmitter, HostListener, Input, NgZone, Output, TemplateRef, ViewChild, ViewContainerRef, ViewEncapsulation } from '@angular/core';
import { Subject } from 'rxjs';
import { Key } from './key';
import { NgMentionsListComponent } from './util/mentions-list.component';
import { applyChangeToValue, escapeRegExp, findStartOfMentionInPlainText, getBoundsOfMentionAtPosition, getCaretPosition, mapPlainTextIndex, markupToRegExp, replacePlaceholders, setCaretPosition, styleProperties } from './util/utils';
/**
 * The Mentions Component
 */
var NgMentionsComponent = /** @class */ (function () {
    function NgMentionsComponent(element, componentResolver, viewContainer, changeDetector, ngZone) {
        this.element = element;
        this.componentResolver = componentResolver;
        this.viewContainer = viewContainer;
        this.changeDetector = changeDetector;
        this.ngZone = ngZone;
        /**
         * The character to trigger the mentions list when a user is typing in the input field
         */
        this.triggerChar = '@';
        /**
         * The markup used to format a mention in the model value
         */
        this.mentionMarkup = '@[__display__](__type__:__id__)';
        /**
         * Optional. When using a custom search (i.e. an API call), the internal searching capability should be disabled.
         */
        this.disableSearch = false;
        /**
         * Only used when internal search is not disabled. This limits the maximum number of items to display in the search
         * result list.
         */
        this.maxItems = -1;
        /**
         * Used to cause the search result list to display in a "drop up" fashion, instead of a typical dropdown.
         */
        this.dropUp = false;
        /**
         * If the supplied mentions are a list of objects, this is the name of the property used to display
         * the mention in the search result list and when formatting a mention in the displayed text.
         */
        this.displayName = 'display';
        /**
         * An event emitted, after the trigger character has been typed, with the user-entered search string.
         */
        this.search = new EventEmitter();
        this.valueChanges = new EventEmitter();
        this.stateChanges = new Subject();
        this.displayContent = '';
        this.lines = [];
        this.highlighterStyle = {};
        this.textAreaClassNames = {};
        this.mentions = [];
        this._value = '';
        this._rows = 1;
        this._columns = 1;
        this.stopSearch = false;
        this._destroyed = new Subject();
        this.newLine = /\n/g;
        this._errorState = false;
    }
    Object.defineProperty(NgMentionsComponent.prototype, "formClass", {
        /**
         * Classes for textarea
         */
        get: function () {
            return Object.keys(this.textAreaClassNames).join(' ');
        },
        set: function (classNames) {
            var _this = this;
            this.textAreaClassNames = {};
            Array.from(classNames.split(' ')).forEach(function (className) {
                _this.textAreaClassNames[className] = true;
            });
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NgMentionsComponent.prototype, "value", {
        get: function () {
            return this._value;
        },
        set: function (value) {
            this.parseLines(value);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NgMentionsComponent.prototype, "required", {
        get: function () {
            return this._required;
        },
        set: function (value) {
            this._required = value;
            this.refreshStyles();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NgMentionsComponent.prototype, "disabled", {
        get: function () {
            return this._disabled;
        },
        set: function (value) {
            this._disabled = value;
            this.refreshStyles();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NgMentionsComponent.prototype, "rows", {
        get: function () {
            return this._rows;
        },
        set: function (value) {
            this._rows = Math.max(1, value);
            this.refreshStyles();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NgMentionsComponent.prototype, "columns", {
        get: function () {
            return this._columns;
        },
        set: function (value) {
            this._columns = Math.max(1, value);
            this.refreshStyles();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NgMentionsComponent.prototype, "mentionItems", {
        /**
         * The list of mentions to display, or filter, in the search result list.
         */
        set: function (value) {
            this.mentions = value;
            if (this.disableSearch && this.mentionsList) {
                this.mentionsList.items = value;
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NgMentionsComponent.prototype, "readonly", {
        get: function () {
            return this.disabled ? 'readonly' : null;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NgMentionsComponent.prototype, "errorState", {
        get: function () {
            return this._errorState;
        },
        enumerable: true,
        configurable: true
    });
    NgMentionsComponent.prototype.ngOnInit = function () {
        this.parseMarkup();
    };
    NgMentionsComponent.prototype.ngOnChanges = function (changes) {
        if ('markup' in changes) {
            this.parseMarkup();
        }
    };
    NgMentionsComponent.prototype.ngAfterViewInit = function () {
        this.parseLines(this._value);
        this.refreshStyles();
    };
    NgMentionsComponent.prototype.ngAfterViewChecked = function () {
        this.refreshStyles();
    };
    NgMentionsComponent.prototype.ngOnDestroy = function () {
        this._destroyed.next();
        this._destroyed.complete();
    };
    NgMentionsComponent.prototype.onWindowResize = function () {
        this.refreshStyles();
    };
    NgMentionsComponent.prototype.onTextAreaScroll = function () {
        this.highlighterElement.nativeElement.scrollTop = this.textAreaInputElement.nativeElement.scrollTop;
    };
    NgMentionsComponent.prototype.onSelect = function (event) {
        this.selectionStart = event.target.selectionStart;
        this.selectionEnd = event.target.selectionEnd;
    };
    NgMentionsComponent.prototype.onChange = function (newPlainTextValue) {
        var _this = this;
        var value = this._value;
        var displayTransform = this.displayTransform.bind(this);
        var selectionStart = this.textAreaInputElement.nativeElement.selectionStart;
        var selectionEnd = this.textAreaInputElement.nativeElement.selectionEnd;
        var bounds = getBoundsOfMentionAtPosition(newPlainTextValue, this.markupSearch, selectionStart, displayTransform);
        if (bounds.start !== -1) {
            newPlainTextValue = newPlainTextValue.substring(0, bounds.start) + newPlainTextValue.substring(bounds.end);
        }
        var newValue = applyChangeToValue(value, this.markupSearch, newPlainTextValue, this.selectionStart, this.selectionEnd, selectionEnd, displayTransform);
        var startOfMention = findStartOfMentionInPlainText(value, this.markupSearch, selectionStart, displayTransform);
        if (startOfMention.start > -1 && this.selectionEnd > startOfMention.start) {
            selectionStart = startOfMention.start;
            selectionEnd = selectionStart;
        }
        this.selectionStart = Math.max(selectionStart, 0);
        this.selectionEnd = Math.max(selectionEnd, 0);
        this.parseLines(newValue);
        if (this.selectionEnd > 0) {
            setTimeout(function () { return setCaretPosition(_this.textAreaInputElement.nativeElement, _this.selectionEnd); });
        }
    };
    NgMentionsComponent.prototype.onKeyDown = function (event) {
        var caretPosition = getCaretPosition(this.textAreaInputElement.nativeElement);
        var characterPressed = event.key;
        var keyCode = event.which || event.keyCode;
        if (!characterPressed) {
            var characterCode = event.which || event.keyCode;
            characterPressed = String.fromCharCode(characterCode);
            if (!event.shiftKey && (characterCode >= 65 && characterCode <= 90)) {
                characterPressed = String.fromCharCode(characterCode + 32);
            }
        }
        if (keyCode === Key.Enter && event.wasSelection && caretPosition < this.startPos) {
            caretPosition = this.startNode.length;
            setCaretPosition(this.startNode, caretPosition);
        }
        var startOfMention = findStartOfMentionInPlainText(this._value, this.markupSearch, caretPosition, this.displayTransform.bind(this));
        if (characterPressed === this.triggerChar) {
            this.setupMentionsList(caretPosition);
        }
        else if (startOfMention.start === -1 && this.startPos >= 0) {
            if (caretPosition <= this.startPos) {
                this.mentionsList.show = false;
                this.startPos = -1;
            }
            else if (keyCode !== Key.Shift && !event.shiftKey && !event.metaKey && !event.altKey && !event.ctrlKey &&
                caretPosition > this.startPos) {
                this.handleKeyDown(event, caretPosition, characterPressed);
            }
        }
        else {
            this.onSelect({ target: this.textAreaInputElement.nativeElement });
        }
    };
    NgMentionsComponent.prototype.onBlur = function (event) {
        if (event instanceof FocusEvent && event.relatedTarget) {
            var element = event.relatedTarget;
            if (element.classList.contains('dropdown-item')) {
                return;
            }
        }
        this.stopEvent(event);
        if (this.mentionsList) {
            this.mentionsList.show = false;
        }
    };
    NgMentionsComponent.prototype.isPartMention = function (part) {
        return typeof part.contents !== 'undefined';
    };
    NgMentionsComponent.prototype.formatMention = function (mention) {
        return this._formatMention(mention.contents);
    };
    NgMentionsComponent.prototype.displayTransform = function () {
        var _ = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            _[_i] = arguments[_i];
        }
        var replaceIndex = this.markupSearch.groups[this.displayName];
        return _[replaceIndex];
    };
    NgMentionsComponent.prototype._formatMention = function (contents) {
        var replaceValue = "$" + this.displayName, replaceIndex;
        var result = contents.replace(this.markupSearch.regEx, replaceValue);
        if (result === replaceValue) {
            replaceIndex = "$" + this.markupSearch.groups[this.displayName];
            result = contents.replace(this.markupSearch.regEx, replaceIndex);
        }
        return result;
    };
    NgMentionsComponent.prototype.stopEvent = function (event) {
        if (event.preventDefault) {
            event.preventDefault();
            event.stopPropagation();
            event.stopImmediatePropagation();
        }
    };
    NgMentionsComponent.prototype.setupMentionsList = function (caretPosition) {
        this.startPos = caretPosition;
        this.startNode = window.getSelection().anchorNode;
        this.searchString = '';
        this.stopSearch = false;
        this.showMentionsList();
        this.updateMentionsList();
    };
    NgMentionsComponent.prototype.handleKeyDown = function (event, caretPosition, characterPressed) {
        var _this = this;
        var keyCode = event.which || event.keyCode;
        if (keyCode === Key.Space) {
            this.startPos = -1;
        }
        else if (keyCode === Key.Backspace && caretPosition > 0) {
            caretPosition--;
            if (caretPosition === this.startPos) {
                this.stopSearch = true;
            }
            this.mentionsList.show = !this.stopSearch;
        }
        else if (this.mentionsList.show) {
            if (keyCode === Key.Tab || keyCode === Key.Enter) {
                this.stopEvent(event);
                this.mentionsList.show = false;
                var value = this._value;
                var start = mapPlainTextIndex(value, this.markupSearch, this.startPos, false, this.displayTransform.bind(this));
                var item = event.item || this.mentionsList.selectedItem;
                var newValue = replacePlaceholders(item, this.markupSearch);
                var newDisplayValue = this._formatMention(newValue);
                caretPosition = this.startPos + newDisplayValue.length;
                var searchString = this.searchString || '';
                value = value.substring(0, start) + newValue + value.substring(start + searchString.length + 1, value.length);
                this.parseLines(value);
                this.startPos = -1;
                this.searchString = '';
                this.stopSearch = true;
                this.mentionsList.show = false;
                this.changeDetector.detectChanges();
                setTimeout(function () {
                    setCaretPosition(_this.textAreaInputElement.nativeElement, caretPosition);
                    _this.onSelect({ target: _this.textAreaInputElement.nativeElement });
                });
                return;
            }
            else if (keyCode === Key.Escape) {
                this.stopEvent(event);
                this.mentionsList.show = false;
                this.stopSearch = true;
                return;
            }
            else if (keyCode === Key.ArrowDown) {
                this.stopEvent(event);
                this.mentionsList.selectNextItem();
                return;
            }
            else if (keyCode === Key.ArrowUp) {
                this.stopEvent(event);
                this.mentionsList.selectPreviousItem();
                return;
            }
            else if (keyCode === Key.Home) {
                this.stopEvent(event);
                this.mentionsList.selectFirstItem();
                return;
            }
            else if (keyCode === Key.End) {
                this.stopEvent(event);
                this.mentionsList.selectLastItem();
                return;
            }
        }
        if (keyCode === Key.ArrowLeft || keyCode === Key.ArrowRight || keyCode === Key.Home || keyCode === Key.End) {
            this.onSelect(event);
            return;
        }
        var mention = this.displayContent.substring(this.startPos + 1, caretPosition);
        if (keyCode !== Key.Backspace) {
            mention += characterPressed;
        }
        this.searchString = mention || '';
        this.updateMentionsList();
    };
    NgMentionsComponent.prototype.getDisplayValue = function (item) {
        if (typeof item === 'string') {
            return item;
        }
        else if (item[this.displayName] !== undefined) {
            return item[this.displayName];
        }
        return null;
    };
    NgMentionsComponent.prototype.showMentionsList = function () {
        var _this = this;
        if (!this.mentionsList) {
            var componentFactory = this.componentResolver.resolveComponentFactory(NgMentionsListComponent);
            var componentRef = this.viewContainer.createComponent(componentFactory);
            this.mentionsList = componentRef.instance;
            this.mentionsList.itemTemplate = this.mentionListTemplate;
            this.mentionsList.displayTransform = this.displayTransform.bind(this);
            this.mentionsList.itemSelected.subscribe(function (item) {
                _this.textAreaInputElement.nativeElement.focus();
                var fakeEvent = { keyCode: Key.Enter, wasSelection: true, item: item };
                _this.onKeyDown(fakeEvent);
            });
            this.mentionsList.displayTransform = this.getDisplayValue.bind(this);
        }
        this.mentionsList.textAreaElement = this.textAreaInputElement.nativeElement;
        this.mentionsList.show = true;
        this.mentionsList.dropUp = this.dropUp;
        this.mentionsList.activeIndex = 0;
        this.mentionsList.position();
        this.ngZone.run(function () { return _this.mentionsList.resetScroll(); });
    };
    NgMentionsComponent.prototype.updateMentionsList = function () {
        var _this = this;
        if (!this.disableSearch) {
            var items = Array.from(this.mentions);
            if (this.searchString) {
                var searchString = this.searchString.toLowerCase(), searchRegEx_1 = new RegExp(escapeRegExp(searchString), 'i');
                items = items.filter(function (item) {
                    var value = _this.getDisplayValue(item);
                    return value !== null && searchRegEx_1.test(value);
                });
                if (this.maxItems > 0) {
                    items = items.slice(0, this.maxItems);
                }
            }
            this.mentionsList.items = items;
        }
        else {
            this.search.emit(this.searchString);
        }
    };
    NgMentionsComponent.prototype.parseMarkup = function () {
        if (this.mentionMarkup.length === 0 || this.mentionMarkup[0] !== this.triggerChar) {
            throw new Error("ng-mentions markup pattern must start with the trigger character \"" + this.triggerChar + "\"");
        }
        this.markupSearch = markupToRegExp(this.mentionMarkup);
    };
    NgMentionsComponent.prototype.parseLines = function (value) {
        var _this = this;
        if (value === void 0) { value = ''; }
        if (value !== this._value) {
            value = value || '';
            var lines = value.split(this.newLine).map(function (line) { return _this.formatMentions(line); });
            var displayContent = lines.map(function (line) { return line.content; }).join('\n');
            if (this.displayContent !== displayContent) {
                this.lines = lines;
                this.displayContent = displayContent;
                this.triggerChange(value);
            }
        }
    };
    NgMentionsComponent.prototype.formatMentions = function (line) {
        var _this = this;
        var lineObj = { originalContent: line, content: line, parts: [] };
        if (line.length === 0) {
            return lineObj;
        }
        var tags = [], mention;
        var regEx = this.markupSearch.regEx;
        regEx.lastIndex = 0;
        while ((mention = regEx.exec(line)) !== null) {
            tags.push({ indices: { start: mention.index, end: mention.index + mention[0].length } });
        }
        var prevTags = [];
        var content = '';
        tags.slice().sort(function (tagA, tagB) { return tagA.indices.start - tagB.indices.start; }).forEach(function (tag) {
            var expectedLength = tag.indices.end - tag.indices.start;
            var contents = line.slice(tag.indices.start, tag.indices.end);
            if (contents.length === expectedLength) {
                var prevIndex = prevTags.length > 0 ? prevTags[prevTags.length - 1].indices.end : 0;
                var before = line.slice(prevIndex, tag.indices.start);
                var partMention = { contents: contents, tag: tag };
                lineObj.parts.push(before);
                lineObj.parts.push(partMention);
                prevTags.push(tag);
                content += before + _this.formatMention(partMention);
            }
        });
        var remainingStart = prevTags.length > 0 ? prevTags[prevTags.length - 1].indices.end : 0;
        var remaining = line.slice(remainingStart);
        lineObj.parts.push(remaining);
        content += remaining;
        lineObj.content = content;
        return lineObj;
    };
    NgMentionsComponent.prototype.refreshStyles = function () {
        var _this = this;
        var element = this.textAreaInputElement.nativeElement;
        var computedStyle = getComputedStyle(element);
        this.highlighterStyle = {};
        styleProperties.forEach(function (prop) {
            _this.highlighterStyle[prop] = computedStyle[prop];
        });
        this.changeDetector.detectChanges();
    };
    NgMentionsComponent.prototype.triggerChange = function (value) {
        this._value = value;
        this.valueChanges.emit(this._value);
        this.changeDetector.detectChanges();
    };
    NgMentionsComponent.decorators = [
        { type: Component, args: [{
                    moduleId: module.id,
                    exportAs: 'ngMentions',
                    selector: 'ng-mentions',
                    template: "\n      <div #highlighter class=\"highlighter\" [ngClass]=\"textAreaClassNames\" [attr.readonly]=\"readonly\"\n           [ngStyle]=\"highlighterStyle\">\n          <div *ngFor=\"let line of lines\">\n              <ng-container *ngFor=\"let part of line.parts\">\n                  <highlighted *ngIf=\"isPartMention(part)\" [tag]=\"part.tag\">{{formatMention(part)}}</highlighted>\n                  <ng-container *ngIf=\"!isPartMention(part)\">{{part}}</ng-container>\n              </ng-container>\n              &nbsp;\n          </div>\n      </div>\n      <textarea\n        #input\n        [rows]=\"rows\"\n        [cols]=\"columns\"\n        [ngModel]=\"displayContent\"\n        [ngClass]=\"textAreaClassNames\"\n        (keydown)=\"onKeyDown($event)\"\n        (blur)=\"onBlur($event)\"\n        (select)=\"onSelect($event)\"\n        (mouseup)=\"onSelect($event)\"\n        (ngModelChange)=\"onChange($event)\"\n        (scroll)=\"onTextAreaScroll()\"\n        [disabled]=\"disabled\"\n        [required]=\"required\"\n        [placeholder]=\"placeholder\"\n      ></textarea>\n  ",
                    styles: [
                        'ng-mentions {position: relative;}',
                        'ng-mentions textarea {position:relative; background-color: transparent !important;}', "ng-mentions .highlighter {\n        position: absolute;\n        top:      0;\n        left:     0;\n        right:    0;\n        bottom:   0;\n        color:    transparent;\n        overflow: hidden !important;\n    }",
                        "ng-mentions highlighted {\n        display:          inline;\n        border-radius:    3px;\n        padding:          1px;\n        margin:           -1px;\n        overflow-wrap:    break-word;\n        background-color: lightblue;\n    }"
                    ],
                    preserveWhitespaces: false,
                    encapsulation: ViewEncapsulation.None
                },] },
    ];
    /** @nocollapse */
    NgMentionsComponent.ctorParameters = function () { return [
        { type: ElementRef },
        { type: ComponentFactoryResolver },
        { type: ViewContainerRef },
        { type: ChangeDetectorRef },
        { type: NgZone }
    ]; };
    NgMentionsComponent.propDecorators = {
        triggerChar: [{ type: Input, args: ['triggerChar',] }],
        mentionMarkup: [{ type: Input, args: ['markup',] }],
        disableSearch: [{ type: Input, args: ['disableSearch',] }],
        maxItems: [{ type: Input, args: ['maxItems',] }],
        dropUp: [{ type: Input, args: ['dropUp',] }],
        displayName: [{ type: Input, args: ['displayName',] }],
        formClass: [{ type: Input, args: ['formClass',] }],
        placeholder: [{ type: Input, args: ['placeholder',] }],
        value: [{ type: Input, args: ['value',] }],
        required: [{ type: Input, args: ['required',] }],
        disabled: [{ type: Input, args: ['disabled',] }],
        rows: [{ type: Input, args: ['rows',] }],
        columns: [{ type: Input, args: ['cols',] }],
        mentionItems: [{ type: Input, args: ['mentions',] }],
        search: [{ type: Output, args: ['search',] }],
        valueChanges: [{ type: Output, args: ['valueChanges',] }],
        stateChanges: [{ type: Output, args: ['stateChanges',] }],
        mentionListTemplate: [{ type: ContentChild, args: [TemplateRef,] }],
        textAreaInputElement: [{ type: ViewChild, args: ['input',] }],
        highlighterElement: [{ type: ViewChild, args: ['highlighter',] }],
        onWindowResize: [{ type: HostListener, args: ['window:resize',] }]
    };
    return NgMentionsComponent;
}());
export { NgMentionsComponent };
//# sourceMappingURL=mentions.component.js.map