export var styleProperties = Object.freeze([
    'direction',
    'boxSizing',
    'width',
    'height', 'overflowX',
    'overflowY',
    'borderTopWidth', 'borderRightWidth', 'borderBottomWidth', 'borderLeftWidth', 'borderStyle',
    'paddingTop', 'paddingRight', 'paddingBottom', 'paddingLeft',
    // https://developer.mozilla.org/en-US/docs/Web/CSS/font
    'fontStyle', 'fontVariant', 'fontWeight', 'fontStretch', 'fontSize', 'fontSizeAdjust', 'lineHeight', 'fontFamily',
    'textAlign', 'textTransform', 'textIndent',
    'textDecoration',
    'letterSpacing', 'wordSpacing',
    'tabSize', 'MozTabSize'
]);
var isBrowser = typeof window !== 'undefined';
var isFirefox = isBrowser && window.mozInnerScreenX != null;
function isInputOrTextAreaElement(element) {
    return element !== null &&
        (element.nodeName === 'INPUT' || element.nodeName === 'TEXTAREA' || element.nodeName === '#text');
}
function spliceString(value, start, end, insert) {
    return value.substring(0, start) + insert + value.substring(end) + '';
}
function iterateMentionsMarkup(value, mentionMarkup, textIterator, markupIterator, displayTransform) {
    var match, start = 0, currentPlainTextIndex = 0;
    var regEx = mentionMarkup.regEx;
    regEx.lastIndex = 0;
    while ((match = regEx.exec(value)) !== null) {
        var display = displayTransform.apply(void 0, match);
        var substr = value.substring(start, match.index);
        textIterator(substr, start, currentPlainTextIndex);
        currentPlainTextIndex += substr.length;
        markupIterator(match[0], match.index, currentPlainTextIndex, display);
        currentPlainTextIndex += display.length;
        start = regEx.lastIndex;
    }
    if (start < value.length) {
        textIterator(value.substring(start), start, currentPlainTextIndex);
    }
}
function iterateOnlyMentionsMarkup(value, mentionMarkup, markupIterator, displayTransform) {
    var match, start = 0, currentPlainTextIndex = 0;
    var regEx = mentionMarkup.regEx;
    regEx.lastIndex = 0;
    while ((match = regEx.exec(value)) !== null) {
        var display = displayTransform.apply(void 0, match);
        var substr = value.substring(start, match.index);
        currentPlainTextIndex += substr.length;
        markupIterator(match[0], match.index, currentPlainTextIndex, display);
        currentPlainTextIndex += display.length;
        start = regEx.lastIndex;
    }
}
export function mapPlainTextIndex(value, mentionMarkup, indexInPlainText, toEndOfMarkup, displayTransform) {
    if (isNaN(indexInPlainText)) {
        return indexInPlainText;
    }
    var result;
    var textIterator = function (matchString, index, substringPlainTextIndex) {
        if (typeof result !== 'undefined') {
            return;
        }
        if (substringPlainTextIndex + matchString.length >= indexInPlainText) {
            result = index + indexInPlainText - substringPlainTextIndex;
        }
    };
    var markupIterator = function (matchString, index, mentionPlainTextIndex, display) {
        if (typeof result !== 'undefined') {
            return;
        }
        if (mentionPlainTextIndex + display.length > indexInPlainText) {
            result = index + (toEndOfMarkup ? matchString.length : 0);
        }
    };
    iterateMentionsMarkup(value, mentionMarkup, textIterator, markupIterator, displayTransform);
    return typeof result !== 'undefined' ? result : value.length;
}
export function getCaretPosition(element) {
    if (isInputOrTextAreaElement(element)) {
        var value = element.value;
        return value.slice(0, element.selectionStart).length;
    }
    var selection = window.getSelection();
    if (selection.rangeCount > 0) {
        var range = selection.getRangeAt(0);
        return range.startOffset;
    }
    return 0;
}
export function getCaretCoordinates(element, position) {
    var coords = { top: 0, left: 0 };
    if (!isBrowser) {
        return coords;
    }
    var div = document.createElement('div');
    document.body.appendChild(div);
    var style = div.style;
    var computed = window.getComputedStyle ? getComputedStyle(element) : element.currentStyle;
    style.whiteSpace = 'pre-wrap';
    if (element.nodeName !== 'INPUT') {
        style.wordWrap = 'break-word';
    }
    style.position = 'absolute';
    style.visibility = 'hidden';
    styleProperties.forEach(function (prop) { return style[prop] = computed[prop]; });
    if (isFirefox) {
        if (element.scrollHeight > parseInt(computed.height, 10)) {
            style.overflowY = 'scroll';
        }
    }
    else {
        style.overflow = 'hidden';
    }
    div.textContent = element.value.substring(0, position);
    if (element.nodeName === 'INPUT') {
        div.textContent = div.textContent.replace(/\s/g, '\u00a0');
    }
    var span = document.createElement('span');
    span.textContent = element.value.substring(position) || '.';
    div.appendChild(span);
    coords = {
        top: span.offsetTop + parseInt(computed['borderTopWidth'], 10),
        left: span.offsetLeft + parseInt(computed['borderLeftWidth'], 10)
    };
    document.body.removeChild(div);
    return coords;
}
export function setCaretPosition(element, position) {
    if (isInputOrTextAreaElement(element) && element.selectionStart !== undefined) {
        element.focus();
        element.setSelectionRange(position, position);
    }
    else {
        var range = document.createRange();
        range.setStart(element, position);
        range.collapse(true);
        var selection = window.getSelection();
        selection.removeAllRanges();
        selection.addRange(range);
    }
}
export function escapeRegExp(str) {
    return str.replace(/[-\/\\^$*+?.()|[\]{}]/g, '\\$&');
}
export function markupToRegExp(markup) {
    var markupPattern = escapeRegExp(markup);
    var placeholderRegExp = /__([\w]+)__/g;
    var placeholders = {};
    var groups = /[\[\]()]+/.test(markup);
    var match, i = groups ? 1 : 0, placeholderExclusion = groups ? '^\\)\\]' : '\\S';
    do {
        match = placeholderRegExp.exec(markupPattern);
        if (match) {
            var placeholder = match[1];
            markupPattern = markupPattern.replace("__" + placeholder + "__", "([" + placeholderExclusion + "]+)");
            placeholders[placeholder] = ++i;
        }
    } while (match);
    return { markup: markup, regEx: new RegExp('(' + markupPattern + ')', 'ig'), groups: placeholders };
}
export function getPlainText(value, mentionMarkup, displayTransform) {
    mentionMarkup.regEx.lastIndex = 0;
    return value.replace(mentionMarkup.regEx, displayTransform);
}
export function replacePlaceholders(item, markupMention) {
    var result = markupMention.markup + '';
    Object.keys(markupMention.groups)
        .forEach(function (key) { return result = result.replace(new RegExp("__" + key + "__", 'g'), typeof item === 'string' ? item : item[key]); });
    return result;
}
export function applyChangeToValue(value, markupMention, plainTextValue, selectionStartBeforeChange, selectionEndBeforeChange, selectionEndAfterChange, displayTransform) {
    if (selectionStartBeforeChange === void 0) { selectionStartBeforeChange = 0; }
    if (selectionEndBeforeChange === void 0) { selectionEndBeforeChange = 0; }
    if (selectionEndAfterChange === void 0) { selectionEndAfterChange = 0; }
    var oldPlainTextValue = getPlainText(value, markupMention, displayTransform);
    var lengthDelta = oldPlainTextValue.length - plainTextValue.length;
    if (!selectionStartBeforeChange) {
        selectionStartBeforeChange = selectionEndBeforeChange + lengthDelta;
    }
    if (!selectionEndBeforeChange) {
        selectionEndBeforeChange = selectionStartBeforeChange;
    }
    if (selectionStartBeforeChange === selectionEndBeforeChange && selectionEndBeforeChange === selectionEndAfterChange &&
        oldPlainTextValue.length === plainTextValue.length) {
        selectionStartBeforeChange--;
    }
    var insert = plainTextValue.slice(selectionStartBeforeChange, selectionEndAfterChange);
    var spliceStart = Math.min(selectionStartBeforeChange, selectionEndAfterChange);
    var spliceEnd = selectionEndBeforeChange;
    if (selectionStartBeforeChange === selectionEndAfterChange) {
        spliceEnd = Math.max(selectionEndBeforeChange, selectionStartBeforeChange + lengthDelta);
    }
    return spliceString(value, mapPlainTextIndex(value, markupMention, spliceStart, false, displayTransform), mapPlainTextIndex(value, markupMention, spliceEnd, true, displayTransform), insert);
}
export function findStartOfMentionInPlainText(value, mentionMarkup, indexInPlainText, displayTransform) {
    var result = { start: -1, end: -1 };
    var markupIterator = function (matchString, index, mentionPlainTextIndex, display) {
        if (mentionPlainTextIndex < indexInPlainText && mentionPlainTextIndex + display.length > indexInPlainText) {
            result = { start: mentionPlainTextIndex, end: mentionPlainTextIndex + display.length };
            return true;
        }
        return false;
    };
    iterateOnlyMentionsMarkup(value, mentionMarkup, markupIterator, displayTransform);
    return result;
}
export function getBoundsOfMentionAtPosition(value, mentionMarkup, indexInPlainText, displayTransform) {
    return findStartOfMentionInPlainText(value, mentionMarkup, indexInPlainText, displayTransform);
}
//# sourceMappingURL=utils.js.map